﻿#region License Information
/* HeuristicLab
 * Copyright (C) 2002-2012 Heuristic and Evolutionary Algorithms Laboratory (HEAL)
 *
 * This file is part of HeuristicLab.
 *
 * HeuristicLab is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * HeuristicLab is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with HeuristicLab. If not, see <http://www.gnu.org/licenses/>.
 */
#endregion

using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using HeuristicLab.Problems.BinPacking.Interfaces;
using HeuristicLab.Problems.BinPacking.PackingItem;
using HeuristicLab.Problems.BinPacking.Shapes;
using HeuristicLab.Persistence.Default.CompositeSerializers.Storable;
using HeuristicLab.Core;
using HeuristicLab.Common;
using HeuristicLab.Parameters;
using HeuristicLab.Data;
using HeuristicLab.Collections;

namespace HeuristicLab.Encodings.PackingEncoding.PackingPlan {
  //RegularPackingPlan defines that the packingItems have to be parallel to the packingBin in every moment
  //SimpleRotation defines that the items may be rotated 
  //  only in one dimension, 
  //  only once and 
  //  only by 90degree
  //    so that parallelism to the correlating bin is ensured
  //This definition implies that the floor of the (3D) objects never changes (what may be a wanted constraint in most cases..)
  [Item("PackingPlan", "Represents a concrete solution for a bin-packing problem.")]
  [StorableClass]
  public class PackingPlan<D, B, I> : ParameterizedNamedItem, IPackingPlan
    where D : class, IPackingDimensions
    where B : PackingShape<D>, IPackingBin
    where I : PackingShape<D>, IPackingItem {

    #region Properties
    public int NrOfBins {
      get {
        if (PackingItemPositions != null)
          return PackingItemPositions.Select(p => p.Value.AssignedBin).Max() + 1;
        else return 0;
      }
    }

    [Storable]
    public ObservableDictionary<int, D> PackingItemPositions { get; set; }
    [Storable]
    public ObservableDictionary<int, B> PackingBinMeasures { get; set; }

    [Storable]
    public ItemList<I> PackingItemMeasures {
      get;
      set;
    }
    [Storable]
    private DoubleValue quality;
    public DoubleValue Quality {
      get { return quality; }
      set {
        if (quality != value) {
          if (quality != null) DeregisterQualityEvents();
          quality = value;
          if (quality != null) RegisterQualityEvents();
          OnQualityChanged();
        }
      }
    }
    #endregion

    public PackingPlan(B binMeasures, ItemList<I> itemMeasures)
      : base() {
      this.PackingItemMeasures = itemMeasures;
      this.PackingBinMeasures = new ObservableDictionary<int, B>();
      this.PackingBinMeasures[0] = binMeasures;
    }
    public PackingPlan(ObservableDictionary<int, B> binMeasures, ItemList<I> itemMeasures)
      : base() {
      this.PackingItemMeasures = itemMeasures;
      this.PackingBinMeasures = binMeasures;
    }

    [StorableConstructor]
    protected PackingPlan(bool deserializing) : base(deserializing) { }
    protected PackingPlan(PackingPlan<D,B,I> original, Cloner cloner)
      : base(original, cloner) {
      PackingItemPositions = new ObservableDictionary<int, D>(original.PackingItemPositions);
      PackingBinMeasures = new ObservableDictionary<int, B>(original.PackingBinMeasures);
      PackingItemMeasures = original.PackingItemMeasures;
    }
    public override IDeepCloneable Clone(Cloner cloner) {
      return new PackingPlan<D,B,I>(this, cloner);
    }

    public B GetPackingBinMeasuresForBinNr(int binNr) {
      if (PackingBinMeasures.ContainsKey(binNr))
        return PackingBinMeasures[binNr];
      else
        return PackingBinMeasures[0];
    }


    #region Events
    public event EventHandler QualityChanged;
    private void OnQualityChanged() {
      var changed = QualityChanged;
      if (changed != null)
        changed(this, EventArgs.Empty);
    }
    private void RegisterQualityEvents() {
      Quality.ValueChanged += new EventHandler(Quality_ValueChanged);
    }
    private void DeregisterQualityEvents() {
      Quality.ValueChanged -= new EventHandler(Quality_ValueChanged);
    }
    private void Quality_ValueChanged(object sender, EventArgs e) {
      OnQualityChanged();
    }

    public event EventHandler PackingItemPositionChanged;
    private void OnPackingItemPositionChanged() {
      var changed = PackingItemPositionChanged;
      if (changed != null)
        changed(this, EventArgs.Empty);
    }
    private void RegisterPackingItemPositionEvents() {
      PackingItemPositions.PropertyChanged += new System.ComponentModel.PropertyChangedEventHandler(PackingItemPosition_PropertyChanged);
    }
    private void DeregisterPackingItemPositionEvents() {
      PackingItemPositions.PropertyChanged -= new System.ComponentModel.PropertyChangedEventHandler(PackingItemPosition_PropertyChanged);
    }
    private void PackingItemPosition_PropertyChanged(object sender, EventArgs e) {
      OnPackingItemPositionChanged();
    }
    #endregion
  }
}
